//_____________________________________________________________________________
//      thread.c, by Stanko Juzbasic
//      this is test/tutorial code for checking/demonstrating basic thread work
//      it takes array length as user argument, gets number of CPUs from system
//      it spawns threads, does some test work on the arrays, then returns
//      on thread joining results are being interpreted and printed
//      some of this paradigm can be useful in computer music number crunching
//      compile: cc -o thread thread.c -lpthread
//_____________________________________________________________________________
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <pthread.h>
#include <sys/sysctl.h>


int nThreads;
char c[32] = {"Computing in thread "};
int size;
//data has to be a global variable!!
int** data;
bool noprint = false;

//_____________________________________________________________________________
// this is OSX/BSD specific:
// similar routines exist for other OS-es
int getCPUnum (void)
{
int numCPU;
int mib[4];
size_t len = sizeof(numCPU); 

// set the mib for hw.ncpu
mib[0] = CTL_HW;
mib[1] = HW_AVAILCPU;  // alternatively, try HW_NCPU;

// get the number of CPUs from the system 
sysctl(mib, 2, &numCPU, &len, NULL, 0);
return numCPU;   
}

//_____________________________________________________________________________
void *work(void *x)
{
  int tid;
  tid = *((int *) x);
  printf("\t%s%d\n",c, tid);
   
  if(data!= NULL){
  for( int i = 0; i < size; i++)
    data[tid][i] = 1 + i%(tid+2);
  }else printf("\t\tdata = NULL\n");
  
  int *ret = calloc(1,sizeof(int));
  *ret = 1;
  return (void*)ret;
}

//_____________________________________________________________________________
int spin_threads(int num)
{
  pthread_t threads[num];
  int thread_args[num];
  int rc, i;
  void* result;
  int thread_sum = 0;
  int* work_sum;
    
  // initialize data
  data = (int**)calloc(num,sizeof(int*));
    work_sum = (int*)calloc(size, sizeof(int));
  // spawn the threads
  for (i=0; i<num; ++i)
    {
      data[i] = (int*)calloc(size, sizeof(int));
      thread_args[i] = i;
      printf("Spawning thread %d\n", i);
      rc = pthread_create(&threads[i], NULL, work, (void *) &thread_args[i]);
    }

  // wait for threads to finish
  for (i=0; i<num; ++i) {
    rc = pthread_join(threads[i], &result);
    thread_sum += *(int*)result;
  }
    
  for (i = 0; i < num; i ++){
      for(int j = 0; j < size; j++){
          work_sum[j] += data[i][j];
        if(!noprint)printf("%d ",data[i][j]);
        }
    if(!noprint)printf("\n");
  }
  printf("_________________________\n");
    
  if(!noprint)for(int j = 0; j < size; j++)
     printf("%d ",work_sum[j]);
  printf("\n");
    
  for (i = 0; i < num; i++)
    free (data[i]);
    
  free (data);
    
return thread_sum;
}
//_____________________________________________________________________________
int main(int argc, char *argv[])
{
  char text[16];
  
  printf("\n-------- THREAD DEMO --------\n");
  
  if( argc > 1)
  if(strcmp("-n",argv[1]) == 0) noprint = true;
    
  printf("\n\tEnter buffer length:\n");
  scanf("%s", text);
  size = (int)atoi(text);
  
  if(size > 20000000)size = 20000000;
    
  nThreads = getCPUnum();
  printf("\nNumber of CPUs: %d\n", nThreads);
  int sum;
  
  sum = spin_threads(nThreads);
    
  printf("Joined %d threads\n\n", sum);

  return 1;
}